// tsc src/Req.ts --declaration & bundle-cli src/Req.ts dist/Req.min.js -b -name=Req -u -udd -udc
const LayerMsg: { success: Function; toast: Function; error: Function } = (self as any).LayerMsg || null;

declare const $: any;

function find<T>(
    predicate: (value: T, index: number, obj: T[]) => boolean,
    thisArg?: ArrayLike<T>,
): T | void {
    const arr = thisArg || this;
    // if (!isArrayLike(arr)) throw new TypeError();
    // if (!isFunction(predicate)) return; // 在typescript中有类型检查，不需要这一句
    const len = arr.length;
    for (let i = 0; i < len; i++) {
        const item: T = arr[i];
        if (predicate(item, i, arr)) return item;
    }
}

function isNaN(target: any): boolean {
    // return String(target) === "NaN"; // "NaN" 会被判断为true
    return "number" === typeof target && target !== target;
}

function includes<T>(thisArg: ArrayLike<T>, searchElement: (v: T, index: number, arr: ArrayLike<T>) => boolean, fromIndex?: number): boolean;
// @overload
function includes<T>(thisArg: ArrayLike<T>, searchElement: T, fromIndex?: number): boolean;
// includes(searchElement: T, fromIndex?: number): boolean
function includes(thisArg, searchElement, fromIndex = 0) {
    const arr = thisArg || this;
    // if (!isArrayLike(arr)) throw new TypeError();
    const len = arr.length;
    for (let i = fromIndex; i < len; i++) {
        const item = arr[i];
        if (typeof searchElement === "function") {
            if (searchElement(item, i, arr)) return true;
        } else {
            if (item === searchElement) return true;
        }
        if (isNaN(item) && isNaN(searchElement)) return true;
    }
    return false;
}

export default class Req {
    ajax: any;
    private ajaxList: any[] = [];
    static readonly default = new Req();

    constructor(
        private config: { baseUrl: string } = {baseUrl: ""},
        private jqAjaxConfig: Record<string, any> = {},
        private codeResolve: { codes: number[], resolve: ((res: any) => Promise<any>) }[] = [
            {
                codes: [200],
                resolve(res) {
                    return Promise.resolve(res);
                }
            }
        ]) {
    }

    request = (method: "get" | "post", url: string, data: object = {}, jqAjaxConfig = this.jqAjaxConfig): Promise<any> => {
        const {codeResolve, config: {baseUrl = ""}} = this;
        const onSuccess = jqAjaxConfig.success;
        delete jqAjaxConfig.success;
        const onError = jqAjaxConfig.error;
        delete jqAjaxConfig.success;
        let ajax: any;

        const promise = new Promise<any>((resolve, reject) => {
            const successCB = (res) => {
                onSuccess && onSuccess(res);
                resolve(res);
            };
            const errorCB = (e) => {
                onError && onError(e);
                reject(e);
            };
            const param = {
                dataType: "json",
                timeout: 10000,
                success(res: any) {
                    const code = Number(res.code ?? res.status);
                    if (!codeResolve || !codeResolve.length) {
                        successCB(res);
                        return;
                    }
                    const item = find((item) => {
                        return includes(item.codes, code);
                    }, codeResolve);
                    if (item) {
                        item.resolve(res).then(successCB, errorCB);
                        return;
                    }
                    LayerMsg?.error(res.msg || res.message);
                    errorCB(res);
                },
                error: function (e) {
                    if (e.statusText !== "abort") {
                        LayerMsg?.error("网络错误，请检查网络后重试");
                    }
                    errorCB(e);
                }
            };
            this.ajax = ajax = $.ajax({
                ...param,
                ...jqAjaxConfig,
                ...{
                    url: baseUrl + url,
                    type: method,
                    data: data,
                }
            });
            this.ajaxList.push(ajax);
        });

        promise.finally(() => {
            const index = this.ajaxList.indexOf(ajax);
            if (index > -1) {
                this.ajaxList.splice(index, 1);
            }
        });

        return promise;
    };

    sync = {
        get: (url: string, data: object = {}, onSuccess?: Function, onError?: Function) => {
            const params = {
                ...this.jqAjaxConfig,
                ...{
                    async: false,
                    success(res) {
                        onSuccess && onSuccess(res);
                    },
                    error(e) {
                        onError && onError(e);
                    }
                }
            };
            return this.request(
                "get",
                url,
                data,
                params
            );
        },
        post: (url: string, data: object = {}, onSuccess?: Function, onError?: Function) => {
            const params = {
                ...this.jqAjaxConfig,
                ...{
                    async: false,
                    success(res) {
                        onSuccess && onSuccess(res);
                    },
                    error(e) {
                        onError && onError(e);
                    }
                }
            };
            return this.request(
                "post",
                url,
                data,
                params
            );
        }
    };

    get = (url: string, data: object = {}) => {
        return this.request("get", url, data);
    };


    post = (url: string, data: object = {}) => {
        return this.request("post", url, data);
    };

    abort = () => {
        this.ajaxList.forEach(aj => {
            aj.abort();
        });
        this.ajaxList = [];
    };

    static sync = Req.default.sync;

    static get = Req.default.get;

    static post = Req.default.post;
}



